#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Input an rt_plan from pydicom into the Plan class to parse MLC data
Created on Wed, Feb 28 2018
@author: Dan Cutright, PhD
"""

import numpy as np
from shapely.geometry import Polygon

MAX_FIELD_SIZE_X = 400  # in mm
MAX_FIELD_SIZE_Y = 400  # in mm


class Plan:
    def __init__(self, rt_plan):
        self.fx_group = [FxGroup(fx_grp_seq, rt_plan.BeamSequence) for fx_grp_seq in rt_plan.FractionGroupSequence]


class FxGroup:
    def __init__(self, fx_grp_seq, plan_beam_sequences):
        self.fxs = fx_grp_seq.NumberOfFractionsPlanned

        meter_set = {}
        for ref_beam in fx_grp_seq.ReferencedBeamSequence:
            ref_beam_num = str(ref_beam.ReferencedBeamNumber)
            meter_set[ref_beam_num] = float(ref_beam.BeamMeterset)

        self.beam = []
        for beam_seq in plan_beam_sequences:
            beam_num = str(beam_seq.BeamNumber)
            if beam_num in meter_set:
                self.beam.append(Beam(beam_seq, meter_set[beam_num]))
        self.beam_count = len(self.beam)


class Beam:
    def __init__(self, beam_seq, meter_set):

        self.meter_set = meter_set

        self.control_point = [ControlPoint(cp_seq) for cp_seq in beam_seq.ControlPointSequence]
        self.control_point_count = len(self.control_point)

        for bld_seq in beam_seq.BeamLimitingDeviceSequence:
            if hasattr(bld_seq, 'LeafPositionBoundaries'):
                lb = bld_seq.LeafPositionBoundaries
                lb.sort(reverse=True)  # Ensure descending order, not sure if a mlcy defaults to that in DICOM
                self.leaf_boundaries = lb

        self.aperture = [get_shapely_from_cp(self.leaf_boundaries, cp) for cp in self.control_point]

        self.control_point_meter_set = np.append([0], np.diff(np.array([cp.cum_mu for cp in self.control_point])))


class ControlPoint:
    def __init__(self, cp_seq):
        cp = {'cum_mu': float(cp_seq.CumulativeMetersetWeight)}
        for device_position_seq in cp_seq.BeamLimitingDevicePositionSequence:
            leaf_jaw_type = str(device_position_seq.RTBeamLimitingDeviceType).lower()

            positions = np.array(map(float, device_position_seq.LeafJawPositions))

            pos_count = len(positions)
            cp[leaf_jaw_type] = [positions[0:pos_count / 2],
                                 positions[pos_count / 2:pos_count]]

        for key in cp:
            setattr(self, key, cp[key])


def get_shapely_from_cp(leaf_boundaries, control_point):
    lb = leaf_boundaries  # NOTE: this is in DESCENDING order
    cp = control_point

    # Determine jaw opening
    if hasattr(cp, 'asymy'):
        y_min = min(cp.asymy)
        y_max = max(cp.asymy)
    else:
        y_min = np.float(-MAX_FIELD_SIZE_Y / 2)
        y_max = np.float(MAX_FIELD_SIZE_Y / 2)
    if hasattr(cp, 'asymx'):
        x_min = min(cp.asymx)
        x_max = max(cp.asymx)
    else:
        x_min = np.float(-MAX_FIELD_SIZE_X / 2)
        x_max = np.float(MAX_FIELD_SIZE_X / 2)

    # Get mlc positions and start/end indices based on jaws
    if hasattr(cp, 'mlcx'):
        mlc = cp.mlcx
        order = 1
    elif hasattr(cp, 'mlcy'):
        mlc = cp.mlcy
        order = -1
    else:
        return False

    v_max = [x_max, y_max][int((order+1)/2)]
    v_min = [x_min, y_min][int((order+1)/2)]

    # These indices are selected to ignore MLCs under the jaws
    # Needs validation, definitely lb_end_index is not tight enough
    lb_start_index = int(np.argmax(lb < v_max)) - 1
    lb_end_index = int(np.argmax(lb < v_min)) - 1

    # Get list of points defining MLC open area
    # To avoid using the intersection function for MLC and Jaw openings
    points = []

    # Accumulate points representing top and bottom positions of A side MLCs
    # First MLC may be split by Jaw
    points.append((mlc[0][lb_start_index], v_max)[::order])
    points.append((mlc[0][lb_start_index], lb[lb_start_index + 1])[::order])
    # The MLCs excluding first and last
    for i in range(lb_start_index+1, lb_end_index):
        points.append((mlc[0][i], lb[i])[::order])
        points.append((mlc[0][i], lb[i+1])[::order])
    # Last MLC may be split by Jaw
    points.append((mlc[0][lb_end_index], lb[lb_end_index])[::order])
    points.append((mlc[0][lb_end_index], v_min)[::order])

    # Accumulate points representing top and bottom positions of B side MLCs, in reverse order
    points.append((mlc[1][lb_end_index], v_min)[::order])
    points.append((mlc[1][lb_end_index], lb[lb_end_index])[::order])
    for i in range(lb_start_index+1, lb_end_index)[::-1]:
        points.append((mlc[1][i], lb[i+1])[::order])
        points.append((mlc[1][i], lb[i])[::order])
    points.append((mlc[1][lb_start_index], lb[lb_start_index + 1])[::order])
    points.append((mlc[1][lb_start_index], v_max)[::order])

    points.append(points[0])  # explicitly close polygon

    aperture = Polygon(points).buffer(0)  # may turn into MultiPolygon

    return aperture



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Mar  2 22:15:52 2017

@author: nightowl
"""

from __future__ import print_function
from sql_connector import DVH_SQL
from dicom_to_python import DVHTable, PlanRow, BeamTable, RxTable
import os
import shutil
import dicom
from datetime import datetime


file_types = {'rtplan', 'rtstruct', 'rtdose'}


def dicom_to_sql(**kwargs): \

    start_time = datetime.now()
    print(str(start_time), 'Beginning import', sep=' ')

    dicom_catalogue_update = []

    # Read SQL configuration file
    script_dir = os.path.dirname(__file__)
    rel_path = "preferences/import_settings.txt"
    abs_file_path = os.path.join(script_dir, rel_path)
    with open(abs_file_path, 'r') as document:
        import_settings = {}
        for line in document:
            line = line.split()
            if not line:
                continue
            import_settings[line[0]] = line[1:][0]
            # Convert strings to boolean
            if line[1:][0].lower() == 'true':
                import_settings[line[0]] = True
            elif line[1:][0].lower() == 'false':
                import_settings[line[0]] = False

    if 'start_path' in kwargs and kwargs['start_path']:
        rel_path = kwargs['start_path']
        abs_file_path = os.path.join(script_dir, rel_path)
        import_settings['inbox'] = abs_file_path

    sqlcnx = DVH_SQL()

    if 'force_update' in kwargs and kwargs['force_update']:
        force_update = True
    else:
        force_update = False

    file_paths = get_file_paths(import_settings['inbox'])

    for uid in list(file_paths):

        if is_uid_imported(uid):
            print("The UID from the following files is already imported.")
            if not force_update:
                print("Must delete content associated with this UID from database before reimporting.")
                print("These files have been moved into the 'misc' folder within your 'imported' folder.")
                for file_type in file_types:
                    print(file_paths[uid][file_type]['file_path'])
                print("The UID is %s" % uid)
                continue

            else:
                print("Force Update set to True. Processing with import.")
                print("WARNING: This import may contain duplicate data already in the database.")

        dicom_catalogue_update.append(uid)

        # Collect and print the file paths
        plan_file = file_paths[uid]['rtplan']['latest_file']
        struct_file = file_paths[uid]['rtstruct']['latest_file']
        dose_file = file_paths[uid]['rtdose']['latest_file']
        print("plan file: %s" % plan_file)
        print("struct file: %s" % struct_file)
        print("dose file: %s" % dose_file)

        # Process DICOM files into Python objects
        plan, beams, dvhs, rxs = [], [], [], []
        mp, ms, md = [], [], []
        if plan_file:
            mp = dicom.read_file(plan_file).ManufacturerModelName.lower()
        if struct_file:
            ms = dicom.read_file(struct_file).ManufacturerModelName.lower()
        if dose_file:
            md = dicom.read_file(dose_file).ManufacturerModelName.lower()

        if 'gammaplan' in "%s %s %s" % (mp, ms, md):
            print("Leksell Gamma Plan is not currently supported. Skipping import.")
            continue

        if plan_file and struct_file and dose_file:
                plan = PlanRow(plan_file, struct_file, dose_file)
        else:
            print('WARNING: Missing complete set of plan, struct, and dose files for uid %s' % uid)
            if not force_update:
                print('WARNING: Skipping this import. If you wish to import an incomplete DICOM set, use Force Update')
                print('WARNING: The current file will be moved to the misc folder with in your imported folder')
                continue

        if plan_file:
            if not hasattr(dicom.read_file(plan_file), 'BrachyTreatmentType'):
                beams = BeamTable(plan_file)
        if struct_file and dose_file:
            dvhs = DVHTable(struct_file, dose_file)
            setattr(dvhs, 'ptv_number', rank_ptvs_by_D95(dvhs))
        if plan_file and struct_file:
            rxs = RxTable(plan_file, struct_file)

        # Insert data from Python objects into SQL tables
        if plan:
            sqlcnx.insert_plan(plan)
        if beams:
            sqlcnx.insert_beams(beams)
        if dvhs:
            sqlcnx.insert_dvhs(dvhs)
        if rxs:
            sqlcnx.insert_rxs(rxs)

        # get mrn for folder name, can't assume a complete set of dose, plan, struct files
        mrn = []
        if dose_file:
            mrn = dicom.read_file(dose_file).PatientID
        elif plan_file:
            mrn = dicom.read_file(plan_file).PatientID
        elif struct_file:
            mrn = dicom.read_file(struct_file).PatientID
        if mrn:
            mrn = "".join(x for x in mrn if x.isalnum())  # remove any special characters
        else:
            mrn = 'NoMRN'

        # convert file_paths[uid] into a list of file paths
        files_to_move = []
        move_types = [i for i in file_types] + ['other']
        for file_type in move_types:
            files_to_move.extend(file_paths[uid][file_type]['file_path'])

        new_folder = os.path.join(import_settings['imported'], mrn)
        move_files_to_new_path(files_to_move, new_folder)

        if plan_file:
            plan_file = os.path.basename(plan_file)
        if struct_file:
            struct_file = os.path.basename(struct_file)
        if dose_file:
            dose_file = os.path.basename(dose_file)

        update_dicom_catalogue(mrn, uid, new_folder, plan_file, struct_file, dose_file)

    # Move remaining files, if any
    move_all_files(import_settings['imported'], import_settings['inbox'])
    remove_empty_folders(import_settings['inbox'])

    sqlcnx.cnx.close()

    end_time = datetime.now()
    print(str(end_time), 'Import complete', sep=' ')

    total_time = end_time - start_time
    seconds = total_time.seconds
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    if h:
        print("This import took %dhrs %02dmin %02dsec to complete" % (h, m, s))
    elif m:
        print("This import took %02dmin %02dsec to complete" % (m, s))
    else:
        print("This import took %02dsec to complete" % s)


def get_file_paths(start_path):
    print('Collecting DICOM file paths')
    f = []
    for root, dirs, files in os.walk(start_path, topdown=False):
        for name in files:
            f.append(os.path.join(root, name))

    # Collect all dicom files by UID, separate non-dicom files into misc
    file_paths = {}
    for file_path in f:
        try:
            dicom_file = dicom.read_file(file_path)
        except:
            dicom_file = False

        if dicom_file:
            uid = dicom_file.StudyInstanceUID
            file_type = dicom_file.Modality.lower()  # (rtplan, rtstruct, rtdose)
            timestamp = os.path.getmtime(file_path)

            if uid not in list(file_paths):
                file_paths[uid] = {'rtplan': {'file_path': [], 'timestamp': [], 'latest_file': []},
                                   'rtstruct': {'file_path': [], 'timestamp': [], 'latest_file': []},
                                   'rtdose': {'file_path': [], 'timestamp': [], 'latest_file': []},
                                   'other': {'file_path': [], 'timestamp': []}}

            if file_type not in file_types:
                file_type = 'other'

            file_paths[uid][file_type]['file_path'].append(file_path)
            file_paths[uid][file_type]['timestamp'].append(timestamp)

    for uid in list(file_paths):
        for file_type in file_types:
            latest_index, latest_time = [], []
            for i, ts in enumerate(file_paths[uid][file_type]['timestamp']):
                if not latest_time or ts > latest_time:
                    latest_index, latest_time = i, ts
            if isinstance(latest_index, int):
                file_paths[uid][file_type]['latest_file'] = file_paths[uid][file_type]['file_path'][latest_index]
            else:
                file_paths[uid][file_type]['latest_file'] = []
    print('DICOM file paths collected')
    return file_paths


def rebuild_database(start_path):
    print('connecting to SQL DB')
    sqlcnx = DVH_SQL()
    print('connection established')

    sqlcnx.reinitialize_database()
    print('DB reinitialized with no data')
    dicom_to_sql(start_path=start_path, force_update=True)
    sqlcnx.cnx.close()


def is_uid_imported(uid):

    for table in {'DVHs', 'Plans', 'Beams', 'Rxs'}:
        if DVH_SQL().is_study_instance_uid_in_table(table, uid):
            return True
    return False


def rank_ptvs_by_D95(dvhs):
    ptv_number_list = [0] * dvhs.count
    ptv_index = [i for i in range(0, dvhs.count) if dvhs.roi_type[i] == 'PTV']

    ptv_count = len(ptv_index)

    # Calculate D95 for each PTV
    doses_to_rank = get_dose_to_volume(dvhs, ptv_index, 0.95)
    order_index = sorted(range(ptv_count), key=lambda k: doses_to_rank[k])
    final_order = sorted(range(ptv_count), key=lambda k: order_index[k])

    for i in range(0, ptv_count):
        ptv_number_list[ptv_index[i]] = final_order[i] + 1

    return ptv_number_list


def get_dose_to_volume(dvhs, indices, roi_fraction):
    # Not precise (i.e., no interpolation) but good enough for sorting PTVs
    doses = []
    for x in indices:
        abs_volume = dvhs.volume[x] * roi_fraction
        dvh = dvhs.dvhs[x]
        dose = next(x[0] for x in enumerate(dvh) if x[1] < abs_volume)
        doses.append(dose)

    return doses


def move_files_to_new_path(files, new_dir):
    for file_path in files:
        file_name = os.path.basename(file_path)
        new = os.path.join(new_dir, file_name)
        try:
            shutil.move(file_path, new)
        except:
            os.mkdir(new_dir)
            shutil.move(file_path, new)


def remove_empty_folders(start_path):
    if start_path[0:2] == './':
        script_dir = os.path.dirname(__file__)
        rel_path = start_path[2:]
        start_path = os.path.join(script_dir, rel_path)

    for (path, dirs, files) in os.walk(start_path, topdown=False):
        if files:
            continue
        try:
            if path != start_path:
                os.rmdir(path)
        except OSError:
            pass


def move_all_files(new_dir, old_dir):
    initial_path = os.path.dirname(os.path.realpath(__file__))

    os.chdir(old_dir)

    file_paths = []
    for path, subdirs, files in os.walk(old_dir):
        for name in files:
            file_paths.append(os.path.join(path, name))

    misc_path = os.path.join(new_dir, 'misc')
    if not os.path.isdir(misc_path):
        os.mkdir(misc_path)

    for f in file_paths:
        file_name = os.path.basename(f)
        new = os.path.join(misc_path, file_name)
        shutil.move(f, new)

    os.chdir(initial_path)


def update_dicom_catalogue(mrn, uid, dir_path, plan_file, struct_file, dose_file):
    if not plan_file:
        plan_file = "(NULL)"
    if not plan_file:
        struct_file = "(NULL)"
    if not plan_file:
        dose_file = "(NULL)"
    DVH_SQL().insert_dicom_file_row(mrn, uid, dir_path, plan_file, struct_file, dose_file)


if __name__ == '__main__':
    DVH_SQL().reinitialize_database()
    dicom_to_sql()



"""
Various options for DVH Analytics
Created on Sat Oct 28 2017
@author: Dan Cutright, PhD
"""

# Setting this to true enables log in screens
# You must add your own code in into check_credentials of auth.py
# main.py sends a usergroup of 'generic', settings.py and admin.py send usergroup of 'admin',
# user this in your check_credentials function to enable proper access
AUTH_USER_REQ = False

# The backup module can be convoluted and is really just a simple GUI for psql dump commands
# if you do not have pgsql command line tool installed, maybe best to just disable this tab to be cautious
DISABLE_BACKUP_TAB = False

# DVH Endpoint Table view cap, bokeh is very slow to delete a table and add another
# This would need to happen if "changing" number of columns in a table, so DVH Analytics prints
# and empty table with the number of endpoints set below.  All endpoints specified in main app will be calulated,
# but only the number specified below will be displayed
ENDPOINT_COUNT = 10

# These colors propagate to all tabs that visualize your two groups
GROUP_1_COLOR = 'blue'
GROUP_1_COLOR_NEG_CORR = 'green'
GROUP_2_COLOR = 'red'
GROUP_2_COLOR_NEG_CORR = 'purple'
GROUP_1_and_2_COLOR = 'purple'  # only for data in time-series that fit both groups

# The line width of selected DVHs in the DVH plot
DVH_LINE_WIDTH = 2

# Adjusts the opacity of the inner-quartile ranges
IQR_1_ALPHA = 0.075
IQR_2_ALPHA = 0.075

# Adjust the plot font sizes
PLOT_AXIS_LABEL_FONT_SIZE = "14pt"
PLOT_AXIS_MAJOR_LABEL_FONT_SIZE = "10pt"

# Number of data points are reduced by this factor during dynamic plot interaction to speed-up visualizations
# This is only applied to the DVH plot since it has a large amount of data
LOD_FACTOR = 100

# Options for the group statistical DVHs in the DVHs tab
STATS_1_MEDIAN_LINE_WIDTH = 1
STATS_1_MEDIAN_LINE_DASH = 'solid'
STATS_1_MEDIAN_ALPHA = 0.6
STATS_1_MEAN_LINE_WIDTH = 2
STATS_1_MEAN_LINE_DASH = 'dashed'
STATS_1_MEAN_ALPHA = 0.5
STATS_2_MEDIAN_LINE_WIDTH = 1
STATS_2_MEDIAN_LINE_DASH = 'solid'
STATS_2_MEDIAN_ALPHA = 0.6
STATS_2_MEAN_LINE_WIDTH = 2
STATS_2_MEAN_LINE_DASH = 'dashed'
STATS_2_MEAN_ALPHA = 0.5

# Options for the time-series plot
TIME_SERIES_1_CIRCLE_SIZE = 10
TIME_SERIES_1_CIRCLE_ALPHA = 0.25
TIME_SERIES_1_TREND_LINE_WIDTH = 1
TIME_SERIES_1_TREND_LINE_DASH = 'solid'
TIME_SERIES_1_AVG_LINE_WIDTH = 1
TIME_SERIES_1_AVG_LINE_DASH = 'dotted'
TIME_SERIES_1_PATCH_ALPHA = 0.1
TIME_SERIES_2_CIRCLE_SIZE = 10
TIME_SERIES_2_CIRCLE_ALPHA = 0.25
TIME_SERIES_2_TREND_LINE_WIDTH = 1
TIME_SERIES_2_TREND_LINE_DASH = 'solid'
TIME_SERIES_2_AVG_LINE_WIDTH = 1
TIME_SERIES_2_AVG_LINE_DASH = 'dotted'
TIME_SERIES_2_PATCH_ALPHA = 0.1

# Adjust the opacity of the histograms
HISTOGRAM_1_ALPHA = 0.3
HISTOGRAM_2_ALPHA = 0.3

# Options for the plot in the Multi-Variable Regression tab
CORRELATION_1_CIRCLE_SIZE = 10
CORRELATION_1_ALPHA = 0.5
CORRELATION_1_LINE_WIDTH = 2
CORRELATION_1_LINE_DASH = 'dashed'
CORRELATION_2_CIRCLE_SIZE = 10
CORRELATION_2_ALPHA = 0.5
CORRELATION_2_LINE_WIDTH = 2
CORRELATION_2_LINE_DASH = 'dashed'



#!/usr/bin/env python

# Users that wish to implement some form of user authentication should fill in this function, keeping the same input
# parameters.  The following example is for LDAP.  Note that proper implementation also necessitates running your
# Bokeh with a reverse proxy implementing SSL with something like Apache or NGINX.  Please see Bokeh documentation for
# some guidance on this. I (the DVH Analytics developer) do not claim to be a security expert by any stretch.  The
# end-user is entirely liable for proper security implementation.

from __future__ import print_function
# import ldap  # This is from pip install python-ldap.  Not necessary, just what we used for our implementaiton.


# Place holder function.  Edit this to implement.
def check_credentials(username, password, usergroup):

    # ############################
    # # Example Code
    # ############################
    # """Verifies credentials for username and password.
    # Returns None on success or a string describing the error on failure
    # # Adapt to your needs
    # """
    # LDAP_SERVER = 'ldaps://someserver:port'
    # # fully qualified AD user name
    # LDAP_USERNAME = '%s@somedomain' % username
    # # your password
    # LDAP_PASSWORD = password
    # base_dn = 'somebasedn'
    # ldap_filter = 'userPrincipalName=%s@somedomain' % username
    # try:
    #     # build a client
    #     ldap_client = ldap.initialize(LDAP_SERVER)
    #
    #     # perform a synchronous bind
    #     ldap_client.set_option(ldap.OPT_REFERRALS, 0)
    #     ldap_client.simple_bind_s(LDAP_USERNAME, LDAP_PASSWORD)
    #
    #     # get all user attributes
    #     result = ldap_client.search_s(base_dn, ldap.SCOPE_SUBTREE, ldap_filter)
    #     ldap_client.unbind()
    # except ldap.INVALID_CREDENTIALS:
    #     ldap_client.unbind()
    #     print("Login attempt: Invalid credentials")
    #     return False
    # except ldap.SERVER_DOWN:
    #     print("Login attempt: Server down")
    #     return False
    # except ldap.LDAPError, e:
    #     ldap_client.unbind()
    #     if type(e.message) == dict and 'desc' in e.message:
    #         print("Login error: %s" % e.message['desc'])
    #         return False
    #     else:
    #         print("Login error: %s" % e.message)
    #         return False
    #     ldap_client.unbind()

    return True
